@@ -1,10 +1,12 @@
 from abc import ABC, abstractmethod
 from dataclasses import dataclass
+from threading import Timer
 from typing import Any, Callable, Iterable, List, Optional, TypedDict, Union
 from i2 import Sig
 from inspect import _empty
 from front.types import BoundData, FrontElementName
 from front.util import deep_merge, get_value
+from i2.signatures import call_forgivingly
 from pydantic import validate_arguments
 
 
@@ -16,11 +18,21 @@ class FrontElementBase(ABC):
     def __post_init__(self):
         self.name = get_value(self.name, self.obj) or ''
 
+    def pre_render(self):
+        pass
+
     @abstractmethod
     def render(self):
-        # raise NotImplementedError('This method needs to be implemented in subclasses.')
         pass
 
+    def post_render(self, render_result):
+        return render_result
+
+    def __call__(self):
+        self.pre_render()
+        r = self.render()
+        return self.post_render(r)
+
 
 ELEMENT_KEY = '_front_element'
 DEFAULT_INPUT_KEY = '_default'
@@ -39,7 +51,7 @@ def mk_element_from_spec(spec: FrontElementSpec):
     return factory(**_spec)
 
 
-def mk_input_element_specs(obj, inputs, stored_value_getter):
+def mk_input_element_specs(obj, inputs):
     def mk_input_spec(p):
         input_spec = inputs_spec.get(p.name, {})
         annot = p.annotation if p.annotation != _empty else None
@@ -50,13 +62,7 @@ def mk_input_element_specs(obj, inputs, stored_value_getter):
         input_spec = deep_merge(type_spec, input_spec)
         dflt_input_key = f'{obj.__name__}_{p.name}'
         input_key = input_spec.get('input_key', dflt_input_key)
-        stored_value = stored_value_getter(input_key) if stored_value_getter else None
-        init_value = (
-            stored_value
-            if stored_value is not None
-            else (p.default if p.default != _empty else None)
-        )
-        return dict(input_spec, obj=p, input_key=input_key, init_value=init_value)
+        return dict(input_spec, obj=p, input_key=input_key)
 
     inputs_spec = dict(inputs)
     default = inputs_spec.pop(DEFAULT_INPUT_KEY, {})
@@ -81,7 +87,7 @@ class FrontContainerBase(FrontElementBase):
 
     def _render_children(self):
         for child in self.children:
-            child.render()
+            child()
 
 
 @dataclass
@@ -107,18 +113,35 @@ class TextSectionBase(FrontComponentBase):
 @dataclass
 class InputBase(FrontComponentBase):
     input_key: str = None
-    init_value: Any = None
-    value: BoundData = None
+    value: Any = None
     on_value_change: Callable[..., None] = None
+    bound_data_factory: Callable = None
+
+    def __post_init__(self):
+        super().__post_init__()
+        if not isinstance(self.value, BoundData):
+            if self.bound_data_factory is None:
+                raise ValueError(
+                    f'No factory provided to build a BoundData instance with id \
+                        {self.input_key}'
+                )
+            value = self.value
+            self.value = self.bound_data_factory(self.input_key)
+            if value is not None:
+                self.value.set(value)
+        if not self.value() and self.obj.default != _empty:
+            self.value.set(self.obj.default)
+
+    def post_render(self, render_result):
+        # self.value.set(render_result)
+        if self.on_value_change:
+            call_forgivingly(self.on_value_change, render_result)
+        return render_result
 
 
 class OutputBase(FrontComponentBase):
     output: Any = None
 
-    def render_output(self, output):
-        self.output = output
-        return self.render()
-
 
 class ExecContainerBase(FrontContainerBase):
     def __init__(
@@ -127,13 +150,10 @@ class ExecContainerBase(FrontContainerBase):
         inputs: dict,
         output: dict,
         name: FrontElementName = None,
-        stored_value_getter: Callable[[str], Any] = None,
         auto_submit: bool = False,
         on_submit: Callable[[Any], None] = None,
     ):
-        element_specs = dict(
-            mk_input_element_specs(obj, inputs, stored_value_getter), output=output
-        )
+        element_specs = dict(mk_input_element_specs(obj, inputs), output=output)
         super().__init__(obj=obj, name=name, **element_specs)
         self.auto_submit = auto_submit
         self.on_submit = on_submit
@@ -146,7 +166,7 @@ class ExecContainerBase(FrontContainerBase):
             or isinstance(child, MultiSourceInputContainerBase)
         ]
         return {
-            input_component.obj.name: input_component.render()
+            input_component.obj.name: input_component()
             for input_component in input_components
         }
 
@@ -156,7 +176,8 @@ class ExecContainerBase(FrontContainerBase):
         output_component = next(
             iter(child for child in self.children if isinstance(child, OutputBase))
         )
-        output_component.render_output(output)
+        output_component.output = output
+        output_component()
         if self.on_submit:
             self.on_submit(output)
 
@@ -167,7 +188,6 @@ class MultiSourceInputContainerBase(FrontContainerBase):
         obj=None,
         name: FrontElementName = None,
         input_key: str = None,
-        init_value: Any = None,
         value: BoundData = None,
         on_value_change: Callable = None,
         **kwargs: FrontElementSpec,
@@ -187,7 +207,7 @@ class MultiSourceInputContainerBase(FrontContainerBase):
 class TextInputBase(InputBase):
     def __post_init__(self):
         super().__post_init__()
-        self.init_value = str(self.init_value) if self.init_value is not None else ''
+        self.value.set(str(self.value()) if self.value() is not None else '')
 
 
 @dataclass
@@ -202,7 +222,7 @@ class IntInputBase(NumberInputBase):
 
     def __post_init__(self):
         super().__post_init__()
-        self.init_value = int(self.init_value) if self.init_value is not None else 0
+        self.value.set(int(self.value()) if self.value() is not None else 0)
 
 
 @dataclass
@@ -213,7 +233,7 @@ class FloatInputBase(NumberInputBase):
 
     def __post_init__(self):
         super().__post_init__()
-        self.init_value = float(self.init_value) if self.init_value is not None else 0.0
+        self.value.set(float(self.value()) if self.value() is not None else 0.0)
 
 
 @dataclass
@@ -228,3 +248,12 @@ class SelectBoxBase(InputBase):
     def __post_init__(self):
         super().__post_init__()
         self.options = self.options or []
+
+    def pre_render(self):
+        self._options = list(
+            (self.options() if callable(self.options) else self.options) or []
+        )
+        value = self.value()
+        self._preselected_index = (
+            self._options.index(value) if value in self._options else 0
+        )
